from typing import get_origin, get_args, Union, Literal, Any, _GenericAlias
from types import NoneType, UnionType, GenericAlias
from adaptivecard._typing import ListLike
from adaptivecard.exceptions import TypeCheckError, ValueCheckError, WrongLiteral, WrongType, WrongTypes, WrongItem, NotListLike
from json import dump


def raise_invalid_pixel_error(arg_name: str, arg_value: str):
    msg = f"argument '{arg_name}' must be numeric or a number ending with 'px', got '{arg_value}' instead"
    raise ValueError(msg)


def convert_to_pixel_string(s):
    if isinstance(s, str):
        if not s.replace('px', '').isdecimal():
            raise ValueError
        s = s.replace('px', '') + 'px'
    elif isinstance(s, int):
        s = str(s) + 'px'
    return s

def is_union(tp):
    """
    Checks if tp is a Union. For unsubscripted Unions, will return False.
    """
    if (origin := get_origin(tp)) is Union or origin is UnionType:
        return True
    return False

def is_type(tp):
    """
    Checks if tp is a type in a broad sense. Anything that can be used as a type hint will be
    considered a type, except Unions and Literal.
    """
    return (isinstance(tp, (type, GenericAlias, _GenericAlias)) and not isinstance(tp, UnionType)) \
    and get_origin(tp) is not Literal

def is_parameterized_type(tp):
    """
    Checks if type is a GenericAlias.
    """
    return isinstance(tp, (GenericAlias, _GenericAlias))

def check_type(arg_name: str | None, arg_value: Any, expected_type):
    """
    Type-checking function, raises TypeError if arg_value doesn't match the expected type.
    """
    if expected_type is Any:
        return

    if is_union(expected_type):
        expected_type = get_args(expected_type)

    if get_origin(expected_type) is Literal:
        if arg_value not in (allowed_values := get_args(expected_type)):
            raise WrongLiteral(arg_name, arg_value, allowed_values)
        return

    elif isinstance(expected_type, tuple):
        for sub_type in expected_type:
            try:
                return check_type(arg_name, arg_value, sub_type)
            except (TypeCheckError, ValueCheckError):
                continue
        else:
            raise WrongTypes(arg_name, arg_value)
    elif is_type(expected_type):
        if is_parameterized_type(expected_type):
            type_origin = get_origin(expected_type)
            type_subscripts = get_args(expected_type)
            check_type(arg_name, arg_value, type_origin)
            container = arg_value
            if type_origin is tuple:
                ... # devo tratar tuplas? Se sim, como?
            else:
                for i, item in enumerate(container):
                    try:
                        check_type(arg_name, item, type_subscripts)
                    except (TypeCheckError, ValueCheckError):
                        raise WrongItem(arg_name, i)
                return
        else:
            if (expected_type is NoneType and arg_value is not None) or \
                (not isinstance(arg_value, expected_type)):
                if expected_type is ListLike:
                    raise NotListLike(arg_name, arg_value)
                raise WrongType(arg_name, arg_value, expected_type)
            return
    else:
        raise Exception("Invalid!")  

def camel_to_snake(s: str):
    l = list(s)
    for i, char in enumerate(l):
        if (lower_char := char.lower()) != char:
            l[i] = "_" + lower_char if i > 0 else lower_char
    return "".join(l)

def snake_to_camel(s: str):
    l = s.split("_")
    for i, word in enumerate(l):
        if i > 0:
            l[i] = word.capitalize()
    return "".join(l)

def save_json(path: str, obj: dict, indent: int = 4):
    if not isinstance(indent, int):
        raise TypeError
    with open(path, 'w') as f:
        dump(obj, f, indent=indent)



from typing import Any, Literal, overload
from adaptivecard._mixin import Mixin
import adaptivecard._base_types as _base_types
from adaptivecard._typing import ListLike, DefaultNone, ElementList




class Content(Mixin):
    """Content é o elemento que recebe o AdaptiveCard e é adicionado à lista atachments, atributo de Message"""
    __slots__ = ("content_type", "content")
    def __init__(self, content: _base_types.AdaptiveCard):
        self.content_type = "application/vnd.microsoft.card.adaptive"
        self.content = content


class Message(Mixin):
    """"Estrutura de mensagem. Um card precisa estar contido em uma mensagem para ser enviado via Teams."""
    __slots__ = ('type', 'attachments')
    def __init__(self, attachments: ListLike[_base_types.Content] = DefaultNone):
        self.type = "message"
        if attachments is DefaultNone:
            attachments = []
        self.attachments = list(attachments)

    def attach(self, content):
        self.attachments.append(content)


class AdaptiveCard(Mixin):
    """O template principal do card"""  # Essas descrições hão de ficar mais detalhadas à medida que eu desenvolver a lib e sua documentação
    __slots__ = ("type", "version", "schema", "body", "actions", "fallback_text", "background_image", "min_height",
                 "rtl", "speak", "lang", "vertical_content_alignment")
    def __init__(self,
                 version: str | float = "1.2",
                 body: _base_types.Element | ListLike[_base_types.Element] = DefaultNone,
                 actions: _base_types.Action | ListLike[_base_types.Action] = DefaultNone,
                 fallback_text: str = DefaultNone,
                 background_image: str | _base_types.BackgroundImage = DefaultNone,
                 min_height: str = DefaultNone,
                 rtl: bool = DefaultNone,
                 speak: str = DefaultNone,
                 lang: str = DefaultNone,
                 vertical_content_alignment: Literal["top", "center", "bottom"] = DefaultNone):
        
        self.type = "AdaptiveCard"
        self.version = version  
        self.schema = "http://adaptivecards.io/schemas/adaptive-card.json"
        if body is DefaultNone:
            body = ElementList()
        self.body: ElementList = body
        self.actions = actions
        self.fallback_text = fallback_text
        self.background_image = background_image
        self.min_height = min_height
        self.rtl = rtl
        self.speak = speak
        self.lang = lang
        self.vertical_content_alignment = vertical_content_alignment

    @property
    def empty(self):
        return len(self.body) == 0

    def append(self, value: _base_types.Element, /):
        if not isinstance(value, _base_types.Element):
            raise TypeError("Can only append an Element")
        self.body.append(value)

    def append_action(self, action: _base_types.Action, /):
        if not isinstance(action, _base_types.Action):
            raise TypeError("Can only append an Action")
        if not hasattr(self, 'actions'):
            self.actions = []
        self.actions.append(action)

    def to_message(self):
        content = Content(content=self)
        msg = Message(attachments=[content])
        return msg

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, k):
        return self.body.__getitem__(k)
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == 'version':
            version = __value
            if isinstance(version, float):
                version = str(version)
            __value = version
        if __name == 'body' and isinstance(__value, _base_types.Element) \
            or __name == 'actions' and isinstance(__value, _base_types.Action):
            __value = ElementList([__value])
        return super().__setattr__(__name, __value)
    

_base_types.AdaptiveCard.register(AdaptiveCard)
_base_types.Content.register(Content)
_base_types.Message.register(Message)



import pytest
from adaptivecard.containers import Container, ColumnSet, Column, Table, TableRow, ActionSet
from adaptivecard.exceptions import *


class Test:
    def test_container(self):
        container = Container()
        with pytest.raises(TypeCheckError):
            Container((1,2,3))
        with pytest.raises(ValueCheckError):
            container.style = "wrong value"
        with pytest.raises(AttributeError):
            container.wrong_attribute = "value"

    def test_column_set(self):
        column_set = ColumnSet()
        with pytest.raises(TypeCheckError):
            column_set.horizontal_alignment = "Left"
        with pytest.raises(ValueError):
            column_set.min_height = "20pc"
        column_set.min_height = 20
        assert column_set.min_height == "20px"

    def test_table(self):
        table = Table()
        table.append([1, 2, 3])
        assert all([isinstance(row, TableRow) for row in table])

    def test_table_row(self):
        table_row = TableRow(cells=[1,2,3])
        assert table_row[2][0].text == '3'
        with pytest.raises(ValueCheckError):
            table_row.style = "wrong style"
    
    def test_action_set(self):
        action_set = ActionSet()
        with pytest.raises(TypeError):
            action_set.append("some text")
        with pytest.raises(AttributeError):
            action_set.wrong_attribute = 2

    def test_to_dict(self):
        container = Container()
        columns = [Column("some text"), Column("some other text")]
        column_set = ColumnSet(columns)
        container.append(column_set)
        assert isinstance(container.to_dict(), dict)


import pytest
from adaptivecard.actions import Submit, ShowCard, Execute, OpenUrl
from adaptivecard.exceptions import *


class Test:
    def test_submit(self):
        with pytest.raises(TypeCheckError):
            Submit((1,2,3), None)
        submit_action = Submit({}, None)
        with pytest.raises(ValueCheckError):
            submit_action.style = "dstructve"
        with pytest.raises(TypeCheckError):
            submit_action.data = []
        assert submit_action.to_dict()["associatedInputs"] == "none"
    def test_show_card(self):
        with pytest.raises(TypeCheckError):
            ShowCard([])


from abc import ABC



class Element(ABC):
    pass


class AdaptiveCard(ABC):
    pass


class ActionSet(ABC):
    pass


class Action(ABC):
    pass


class ShowCard(ABC):
    pass


class Execute(ABC):
    pass


class OpenUrl(ABC):
    pass


class Submit(ABC):
    pass


class ToggleVisibility(ABC):
    pass


class TableRow(ABC):
    pass


class TableCell(ABC):
    pass


class Content(ABC):
    pass


class Message(ABC):
    pass


class Container(ABC):
    pass


class ColumnSet(ABC):
    pass


class Column(ABC):
    pass


class Table(ABC):
    pass


class TextBlock(ABC):
    pass


class Image(ABC):
    pass


class Text(ABC):
    pass


class Number(ABC):
    pass


class Date(ABC):
    pass


class Time(ABC):
    pass


class DataQuery(ABC):
    pass


class Toggle(ABC):
    pass


class Choice(ABC):
    pass


class ChoiceSet(ABC):
    pass


class BackgroundImage(ABC):
    pass


Element.register(AdaptiveCard)
Element.register(Table)
Element.register(TableRow)
Element.register(TableCell)
Element.register(ColumnSet)
Element.register(Column)
Element.register(Container)
Element.register(TextBlock)
Element.register(Image)
Element.register(Message)
Element.register(Content)
Element.register(Action)
Element.register(ActionSet)
Element.register(Text)
Element.register(Number)
Element.register(Date)
Element.register(Time)
Element.register(DataQuery)
Element.register(Toggle)
Element.register(Choice)
Element.register(ChoiceSet)
Action.register(ShowCard)
Action.register(Execute)
Action.register(OpenUrl)
Action.register(Submit)
Action.register(ToggleVisibility)




from typing import Literal
from adaptivecard._typing import DefaultNone
import adaptivecard._base_types
from adaptivecard._mixin import Mixin


class Backgroundimage(Mixin):
    __slots__ = ("url", "fill_mode", "horizontal_alignment", "vertical_alignment")
    def __init__(self,
                 url: str,
                 fill_mode: Literal["cover", "repeatHorizontally",
                                    "repeatVertically", "repeat"] = DefaultNone,
                 horizontal_alignment: Literal["left", "center", "right"] = DefaultNone,
                 vertical_alignment: Literal["top", "center", "bottom"] = DefaultNone
                 ) -> None:
        self.url = url
        self.fill_mode = fill_mode
        self.horizontal_alignment = horizontal_alignment
        self.vertical_alignment = vertical_alignment



adaptivecard._base_types.BackgroundImage.register(Backgroundimage)


import pytest
from adaptivecard.types import Backgroundimage
from adaptivecard.exceptions import *


class Test:
    def test_background_image(self):
        url = "https://images.ctfassets.net/sfnkq8lmu5d7/4Ma58uke8SXDQLWYefWiIt/3f1945422ea07ea6520c7aae39180101/2021-11-24_Singleton_Puppy_Syndrome_One_Puppy_Litter.jpg?w=1000&h=750&fl=progressive&q=70&fm=jpg"
        bg_image = Backgroundimage(url=url)
        with pytest.raises(ValueCheckError):
            bg_image.fill_mode = "covre"
        bg_image.fill_mode = "cover"
        bg_image.horizontal_alignment = "center"
        bg_image.to_dict()



import pytest
from adaptivecard.inputs import Text, Number, Date, Time, Toggle, Choice, ChoiceSet
from adaptivecard.exceptions import *


class Test:
    def test_text(self):
        input_text = Text(id="some id", is_multiline=True)
        with pytest.raises(TypeCheckError):
            input_text.spacing = "pading"
        assert isinstance(input_text.to_dict(), dict)

    def test_number(self):
        input_number = Number(id="some id")
        with pytest.raises(TypeCheckError):
            input_number.max = 3.4
        input_number.is_visible = True
        assert isinstance(input_number.to_dict(), dict)

    def test_date(self):
        input_date = Date(id="some id", max="2022-01-05", label="some label")

    def test_time(self):
        input_time = Time(id="some id", height="stretch", placeholder="some placeholder")
        with pytest.raises(TypeCheckError):
            input_time.fallback = 2
        input_time.spacing = None
        assert isinstance(input_time.to_dict(), dict)
    
    def test_toggle(self):
        input_toggle = Toggle(title="some title", id="some id")
        with pytest.raises(TypeCheckError):
            input_toggle.value = 3

    def test_choice_set(self):
        input_choice_set = ChoiceSet(id="some id", wrap=True, label="some lable")
        input_choice_set.choices = Choice(title="some title", value="some value")
        assert isinstance(input_choice_set.choices, list)


from typing import Any, get_type_hints
from adaptivecard._typing import DefaultNone
from adaptivecard._base_types import Element, BackgroundImage
from adaptivecard._utils import check_type, snake_to_camel


class Mixin:
    __slots__ = ()

    def to_dict(self):
        dic = {}
        for attr_name, attr_value in {attr_name: getattr(self, attr_name) for attr_name in self.__slots__ if hasattr(self, attr_name)}.items():
            camel_formated_attr_name = snake_to_camel(attr_name)
            if isinstance(attr_value, (Element, BackgroundImage)):
                dic[camel_formated_attr_name] = attr_value.to_dict()
            elif isinstance(attr_value, list):
                dic[camel_formated_attr_name] = [inner_value.to_dict() for inner_value in attr_value if hasattr(inner_value, "__slots__")]
            else:
                attr_value = attr_value if attr_value is not None else "none"
                dic[camel_formated_attr_name] = attr_value
        return dic

    def __setattr__(self, __name: str, __value: Any) -> None:
        # do not create attributes that are set to DefaultNone, i.e., the user did not input any value
        if __value is DefaultNone:
            return
        type_hints = get_type_hints(self.__init__)
        check_type(__name, __value, type_hints.get(__name, Any))
        super().__setattr__(__name, __value)



from typing import Any, Literal
import adaptivecard._base_types as _base_types
from adaptivecard._typing import DefaultNone
from adaptivecard._mixin import Mixin
from adaptivecard._utils import convert_to_pixel_string, raise_invalid_pixel_error



class TextBlock(Mixin):
    """Elemento de texto"""
    __slots__ = ('type', 'text', 'color', 'font_type', 'horizontal_alignment', 'is_subtle',
                 'max_lines', 'size', 'weight', 'wrap', 'style', 'fallback',
                 'height', 'separator', 'spacing', 'id', 'is_visible')
    def __init__(self,
                 text: Any,
                 color: Literal["default", "dark", "light", "accent", "good", "warning", "attention"] = DefaultNone,
                 font_type: Literal["default", "monospace"] = DefaultNone,
                 horizontal_alignment: Literal["left", "center", "right"] = DefaultNone,
                 is_subtle: bool = DefaultNone,
                 max_lines: int = DefaultNone,
                 size: Literal["default", "small", "medium", "large", "extraLarge"] = DefaultNone,
                 weight: Literal["default", "lighter", "bolder"] = DefaultNone,
                 wrap: bool = DefaultNone,
                 style: Literal["default", "heading"] = DefaultNone,
                 fallback: str | _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge", "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "TextBlock"
        self.text = text
        self.color = color
        self.font_type = font_type
        self.horizontal_alignment = horizontal_alignment
        self.is_subtle = is_subtle
        self.max_lines = max_lines
        self.size = size
        self.weight = weight
        self.wrap = wrap
        self.style = style
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    def __repr__(self):
        return f"{self.__class__.__name__}(text='{self.text}')"

    def __str__(self):
        return self.text

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == 'text' and isinstance(__name, str):
            __value = str(__value)
        return super().__setattr__(__name, __value)


class Image(Mixin):
    __slots__ = ("type", "url", "alt_text", "background_color", "height", "horizontal_alignment",
                 "select_action", "size", "style", "width", "fallback", "separator", "spacing",
                 "id", "is_visible")
    def __init__(self,
                 url: str,
                 alt_text: str = DefaultNone,
                 background_color: str = DefaultNone,
                 height: int | str | Literal["auto", "stretch"] = DefaultNone,
                 horizontal_alignment: Literal["left", "center", "right"] = DefaultNone,
                 select_action: _base_types.Execute | _base_types.OpenUrl | _base_types.Submit |
                 _base_types.ToggleVisibility = DefaultNone,
                 size: Literal["auto", "stretch", "small", "medium", "large"] = DefaultNone,
                 style: Literal["default", "person"] = DefaultNone,
                 width: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone
                ):

        self.type = "Image"
        self.url = url
        self.alt_text = alt_text
        self.background_color = background_color
        self.height = height
        self.horizontal_alignment = horizontal_alignment
        self.select_action = select_action
        self.size = size
        self.style = style
        self.width = width
        self.fallback = fallback
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "width":
            width = __value
            try:
                width = convert_to_pixel_string(width)
            except ValueError:
                raise_invalid_pixel_error(__name, width)
            __value = width
        elif __name == "height":
            if __value not in ["auto", "stretch"]:
                if isinstance(__value, (str, int)):
                    try:
                        __value = convert_to_pixel_string(__value)
                    except ValueError:
                        raise_invalid_pixel_error()
        return super().__setattr__(__name, __value)


_base_types.TextBlock.register(TextBlock)
_base_types.Image.register(Image)


class ValueCheckError(ValueError):
    def __init__(self, message: str = ""):
        super().__init__(message)

class TypeCheckError(TypeError):
    def __init__(self, message: str = "") -> None:
        super().__init__(message)

class WrongLiteral(ValueCheckError):
    def __init__(self, arg_name: str = None, arg_value=None, allowed_values: list | tuple = None):
        if arg_name is None or allowed_values is None:
            super().__init__()
        else:
            msg = f"Argument '{arg_name}' ({arg_value}) must match one of the following values: {', '.join(allowed_values)}"
            super().__init__(msg)

class WrongType(TypeCheckError):
    """
    Raised when type does not match expected type.
    """
    def __init__(self, arg_name: str = None, arg_value: str = None, expected_type = None):
        if arg_name is None or arg_value is None:
            super().__init__()
        else:
            msg = f"Expected argument '{arg_name}' ({arg_value}) to be an instance of {expected_type.__name__}, got {type(arg_value).__name__} instead"
            super().__init__(msg)

class WrongTypes(TypeCheckError):
    """
    Raised when type does not match any of the types in a union.
    """
    def __init__(self, arg_name: str = None, arg_value: str = None):
        if arg_name is None or arg_value is None:
            super().__init__()
        else:
            msg = f"Argument '{arg_name}' ({arg_value}) of type {type(arg_value).__name__} does not match any of the allowed types"
            super().__init__(msg)

class WrongItem(TypeCheckError):
    """
    Raised when the items in a list-like value do not all match the expected types.
    """
    def __init__(self, arg_name: str = None, idx: int = None):
        if arg_name is None or idx is None:
            super().__init__()
        else:
            msg = f"Item {idx} of {arg_name}"
            super().__init__(msg)

class NotListLike(TypeCheckError):
    """
    Raised when value is not of a list-like type.
    """
    def __init__(self, arg_name: str = None, arg_value: str = None) -> None:
        if arg_name is None or arg_value is None:
            super().__init__()
        else:
            msg = f"Argument '{arg_name}' ({arg_value}) must be a collection of some kind, got {type(arg_value).__name__} instead"
            super().__init__(msg)




import pytest
from adaptivecard import AdaptiveCard
from adaptivecard.card_elements import TextBlock
from adaptivecard.containers import Container
from adaptivecard.exceptions import *


class Test:
    def test_(self):
        card = AdaptiveCard()
        with pytest.raises(TypeCheckError):
            AdaptiveCard([1,2,3])
        with pytest.raises(TypeError):
            card.append("")
        with pytest.raises(TypeError):
            card.append_action("")
        with pytest.raises(TypeError):
            card.body.append(1)
        with pytest.raises(AttributeError):
            card.wrong_attribute = "value"
    def test_to_dict(self):
        card = AdaptiveCard(body=Container(TextBlock("some text"))).to_dict()
        assert isinstance(card, dict)
    def test__(self):
        assert AdaptiveCard().empty


from typing import Any, Literal
import adaptivecard._base_types as _base_types
from adaptivecard._typing import ListLike, DefaultNone, ChoiceList
from adaptivecard._mixin import Mixin


class Text(Mixin):
    __slots__ = ("type", "id", "is_multiline", "max_length", "placeholder",
                 "regex", "style", "inline_action", "value", "wrap", "error_message",
                 "is_required", "label", "fallback", "height", "separator", "spacing", "is_visible")

    def __init__(self,
                 id: str,
                 is_multiline: bool = DefaultNone,
                 max_length: int = DefaultNone,
                 placeholder : str = DefaultNone,
                 regex: str = DefaultNone,
                 style: str = DefaultNone,
                 inline_action : _base_types.Action = DefaultNone,
                 value: str = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone,
                 ):
        self.type = "Input.Text"
        self.id = id
        self.is_multiline = is_multiline
        self.max_length = max_length
        self.placeholder = placeholder
        self.regex = regex
        self.style = style
        self.inline_action = inline_action
        self.value = value
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible


class Number(Mixin):
    __slots__ = ("type", "id", "max", "min", "placeholder",
                 "value", "wrap", "error_message", "is_required", "label", "fallback",
                 "height", "separator", "spacing", "is_visible")

    def __init__(self,
                 id: str,
                 max: int = DefaultNone,
                 min: int = DefaultNone,
                 placeholder: str = DefaultNone,
                 value: int = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone,
                 ):

        self.type = "Input.Number"
        self.id = id
        self.max = max
        self.min = min
        self.placeholder = placeholder
        self.value = value
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible


class Date(Mixin):
    __slots__ = ("type", "id", "max", "min", "placeholder", "value", "error_message", "is_required",
                 "label", "fallback", "height", "separator", "spacing", "is_visible")
    def __init__(self,
                 id: str,
                 max: str = DefaultNone,
                 min: str = DefaultNone,
                 placeholder: str = DefaultNone,
                 value: str = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone,) -> None:

        self.type = "Input.Date"
        self.id = id
        self.max = max
        self.min = min
        self.placeholder = placeholder
        self.value = value
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible


class Time(Mixin):
    __slots__ = ("type", "id", "max", "min", "placeholder", "value", "error_message", "is_required",
                 "label", "fallback", "height", "separator", "spacing", "is_visible")
    def __init__(self,
                 id: str,
                 max: str = DefaultNone,
                 min: str = DefaultNone,
                 placeholder: str = DefaultNone,
                 value: str = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "Input.Time"
        self.id = id
        self.max = max
        self.min = min
        self.placeholder = placeholder
        self.value = value
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible        


class DataQuery(Mixin):
    __slots__ = ("type", "dataset", "count", "skip")
    def __init__(self,
                 dataset: str,
                 count: int = DefaultNone,
                 skip: int = DefaultNone
                 ):

        self.type = "Data.Query"
        self.dataset = dataset
        self.count = count
        self.skip = skip


class Toggle(Mixin):
    __slots__ = ('type', 'title', 'id', 'value', 'value_off', 'value_on', 'wrap', 'error_message',
                 'is_required', 'label', 'fallback', 'height', 'separator', 'spacing', 'is_visible')
    def __init__(self,
                 title: str,
                 id: str,
                 value: bool = DefaultNone,
                 value_off: bool = DefaultNone,
                 value_on: bool = DefaultNone,
                 wrap: bool = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "Input.Toggle"
        self.title = title
        self.id = id
        self.value = value
        self.value_off = value_off
        self.value_on = value_on
        self.wrap = wrap
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible  


class Choice(Mixin):
    __slots__ = ("title", "value")
    def __init__(self,
                 title: str,
                 value: str
                 ):
        self.title = title
        self.value = value


class ChoiceSet(Mixin):
    __slots__ = ("type", "id", "choices", "choices_data", "is_multiselect", "style", "value",
                 "placeholder", "wrap", "error_message", "is_required", "label", "fallback",
                 "height", "separator", "spacing", "is_visible")
    def __init__(self,
                 id: str,
                 choices: Choice| ListLike[Choice] = DefaultNone,
                 choices_data: DataQuery = DefaultNone,
                 is_multiselect: bool = DefaultNone,
                 style: Literal["compact", "extended", "filtered"] = DefaultNone,
                 value: str = DefaultNone,
                 placeholder: str = DefaultNone,
                 wrap: bool = DefaultNone,
                 error_message: str = DefaultNone,
                 is_required : bool = DefaultNone,
                 label: str = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 is_visible: bool = DefaultNone,
                 ):

        self.type = "Input.ChoiceSet"
        self.id = id
        if choices is DefaultNone:
            choices = ChoiceList()
        self.choices: ChoiceList = choices
        self.choices_data = choices_data
        self.is_multiselect = is_multiselect
        self.style = style
        self.value = value
        self.placeholder = placeholder
        self.wrap = wrap
        self.error_message = error_message
        self.is_required = is_required
        self.label = label
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.is_visible = is_visible
    
    def append(self, choice: Choice):
        self.choices.append(choice)

    def to_dict(self):
        dic = super().to_dict()
        dic = {
            key: value if key != "choices_data"
            else "choices.data"
            for key, value in dic.items()
        }
        return dic

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "choices":
            if isinstance(__value, Choice):
                __value = ChoiceList([__value])
            elif isinstance(__value, ListLike) and not isinstance(__value, ChoiceList):
                __value = ChoiceList(__value)
            
        return super().__setattr__(__name, __value)
    

_base_types.Text.register(Text)
_base_types.Number.register(Number)
_base_types.Date.register(Date)
_base_types.Time.register(Time)
_base_types.DataQuery.register(DataQuery)
_base_types.Toggle.register(Toggle)
_base_types.Choice.register(Choice)
_base_types.ChoiceSet.register(ChoiceSet)





from typing import Any, Literal, overload
from sys import maxsize
import adaptivecard._base_types as _base_types
from adaptivecard._mixin import Mixin
from adaptivecard.card_elements import TextBlock
from adaptivecard._utils import convert_to_pixel_string, raise_invalid_pixel_error
from adaptivecard._typing import ListLike, DefaultNone, ElementList, ColumnList, RowList
from tabulate import tabulate



class Container(Mixin):
    """A grouping of elements. Containers are useful for grouping a number of related elements
    into one structure. All elements inside a container will inherit its styling attributes
    upon rendering of the card."""
    __slots__ = ('type', 'items', 'style', 'vertical_content_alignment', 'bleed', 'min_height',
                 'rtl', 'height', 'separator', 'spacing', 'id', 'is_visible')
    def __init__(self,
                 items: _base_types.Element | ListLike[_base_types.Element] = DefaultNone,
                 style: Literal["default", "emphasis", "good", "attention", "warning", "accent"] = DefaultNone,
                 vertical_content_alignment: Literal["top", "center", "bottom"] = DefaultNone,
                 bleed: bool = DefaultNone,
                 min_height: str | int = DefaultNone,
                 rtl: bool = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "Container"
        if items is DefaultNone:
            items = ElementList()
        self.items: ElementList = items
        self.style = style
        self.vertical_content_alignment = vertical_content_alignment
        self.bleed = bleed
        self.min_height = min_height
        self.rtl = rtl
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    @property
    def empty(self):
        return len(self.items) == 0

    def append(self, element: _base_types.Element):
        self.items.append(element)

    def __iter__(self):
        return iter(self.items)

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, k):
        return self.items.__getitem__(k)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(items={self.items})"

    def __str__(self) -> str:
        return "[" + ", ".join([str(item) for item in self.items]) + "]"
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == 'items':
            if isinstance(__value, ListLike):
                __value = ElementList(__value)
            elif isinstance(__value, _base_types.Element):
                __value = ElementList([__value])
        elif __name == "min_height":
            try:
                __value = convert_to_pixel_string(__value)
            except ValueError:
                raise_invalid_pixel_error()
        return super().__setattr__(__name, __value)


class Column(Mixin):
    """A column container. Columns must be grouped inside a ColumnSet."""
    __slots__ = ('type', 'items', 'background_image', 'bleed', 'fallback', 'min_height',
                 'rtl', 'separator', 'spacing', 'style', 'vertical_content_alignment', 'rtl',
                 'width', 'id', 'is_visible')
    def __init__(self,
                 items: _base_types.Element | ListLike[_base_types.Element | Any] | Any = DefaultNone,
                 background_image: str | _base_types.BackgroundImage = DefaultNone,
                 bleed: bool = DefaultNone,
                 fallback: "Column" = DefaultNone,
                 min_height: str | int = DefaultNone,
                 rtl: bool = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 select_action: _base_types.Execute | _base_types.OpenUrl | _base_types.Submit
                   | _base_types.ToggleVisibility = DefaultNone,
                 style: Literal["default", "emphasis", "good", "attention", "warning",
                                "accent"] = DefaultNone,
                 vertical_content_alignment: Literal["top", "center", "bottom"] = DefaultNone,
                 width: str | int = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "Column"
        if items is DefaultNone:
            items = ElementList()
        self.items: ElementList = items
        self.background_image = background_image
        self.bleed = bleed
        self.fallback = fallback
        self.min_height = min_height
        self.rtl = rtl
        self.separator = separator
        self.spacing = spacing
        self.select_action = select_action
        self.style = style
        self.vertical_content_alignment = vertical_content_alignment
        self.width = width
        self.id = id
        self.is_visible = is_visible

    def append(self, value: _base_types.Element | Any):
        if not isinstance(value, _base_types.Element):
            value = TextBlock(value)
        self.items.append(value)

    def __iter__(self):
        return iter(self.items)
    
    def __getitem__(self, __i, /):
        return self.items.__getitem__(__i)
    
    def __setitem__(self, __key, __value, /):
        return self.items.__setitem__(__key, __value)
    
    def __str__(self):
        return str([str(item) for item in self.items])
    
    def __repr__(self):
        return f"{self.__class__.__name__}({self.items})"

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "items":
            if isinstance(__value, _base_types.Element):
                __value = ElementList([__value])
            elif isinstance(__value, ListLike):
                __value = ElementList([item 
                                 if isinstance(item, _base_types.Element) 
                                 else TextBlock(item)
                                 for item in __value])
            else:
                __value = ElementList([TextBlock(__value)])
        elif __name == "min_height" or __name == "width":
            try:
                __value = convert_to_pixel_string(__value)
            except ValueError:
                raise_invalid_pixel_error(__name, __value)
        return super().__setattr__(__name, __value)


class ColumnSet(Mixin):
    """A container for columns."""
    __slots__ = ('type', 'columns', 'style', 'bleed', 'min_height', 'horizontal_alignment', 'height',
                 'separator', 'spacing', 'id', 'is_visible')
    def __init__(self,
                 columns: ListLike[Column | ListLike[Any]] = DefaultNone,
                 select_action: _base_types.Execute | _base_types.OpenUrl | _base_types.Submit
                   | _base_types.ToggleVisibility = DefaultNone,
                 style: Literal["default", "emphasis", "good", "attention", "warning",
                                "accent"] = DefaultNone,
                 bleed: bool = DefaultNone,
                 min_height: str | int = DefaultNone,
                 horizontal_alignment: Literal["left", "center", "right"] | None = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] | None = DefaultNone,
                 separator: bool | None = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = 'ColumnSet'
        if columns is DefaultNone:
            columns = ColumnList()
        self.columns: ColumnList = columns
        self.style = style
        self.bleed = bleed
        self.min_height = min_height
        self.horizontal_alignment = horizontal_alignment
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    def append(self, column: Column | ListLike):
        if not isinstance(column, (Column, ListLike)):
            raise TypeError(f"Expected Column or list-like object, got {type(column).__name__} instead")
        if not isinstance(column, Column):
            column = Column(column)
        self.columns.append(column)

    def __iter__(self):
        return iter(self.columns)

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, __k, /):
        return self.columns.__getitem__(__k)
    
    def __setitem__(self, __key: int, __value: object, /):
        if not isinstance(__value, Column):
            raise TypeError()
        return self.columns.__setitem__(__key, __value)
    
    def __repr__(self):
        return f"{self.__class__.__name__}({self.columns})"
    
    def __str__(self):
        return "[" + ", ".join([str(col) for col in self.columns]) + "]"

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "columns":
            if not isinstance(__value, ListLike):
                __value = [__value]
            columns = []
            for col in __value:
                if not isinstance(col, _base_types.Element):
                    col = TextBlock(col)
                if not isinstance(col, Column):
                    col = Column(col)
                columns.append(col)
            __value = ColumnList(columns)
        elif __name == "min_height":
            try:
                __value = convert_to_pixel_string(__value)
            except ValueError:
                raise_invalid_pixel_error(__name, __value)
        return super().__setattr__(__name, __value)


class TableCell(Mixin):
    __slots__ = ('type', 'items', 'select_action', 'style', 'vertical_alignment', 'bleed',
                 'background_image', 'min_height', 'rtl')
    def __init__(self,
                 items: Any | ListLike[Any] = DefaultNone,
                 select_action: _base_types.Execute | _base_types.OpenUrl | _base_types.Submit
                 | _base_types.ToggleVisibility = DefaultNone,
                 style: Literal["default", "emphasis", "good", "attention", "warning",
                                "accent"] = DefaultNone,
                 vertical_alignment: Literal["top", "center", "bottom"] = DefaultNone,
                 bleed: bool = DefaultNone,
                 background_image: str | _base_types.BackgroundImage = DefaultNone,
                 min_height: str | int = DefaultNone,
                 rtl: bool = DefaultNone):

        self.type = "TableCell"
        if items is DefaultNone:
            items = ElementList()
        self.items: ElementList = items
        self.select_action = select_action
        self.items = items
        self.style = style
        self.vertical_alignment = vertical_alignment
        self.bleed = bleed
        self.background_image = background_image
        self.min_height = min_height
        self.rtl = rtl

    def append(self, item: _base_types.Element | Any):
        if not isinstance(item, _base_types.Element):
            item = TextBlock(item)
        self.items.append(item)

    def squeeze(self):
        ...

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, __k, /):
        return self.items.__getitem__(__k)
    
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, _base_types.Element):
            raise TypeError
        return self.items.__setitem__(__key, __value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(items={[item for item in self.items]})"

    def __str__(self) -> str:
        return "[" + ", ".join([str(item) for item in self.items]) + "]"

    def __iter__(self):
        return iter(self.items)
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "items":
            items = __value
            if isinstance(items, _base_types.Element):
                items = ElementList([items])
            elif isinstance(items, ListLike):
                if not isinstance(items, ElementList):
                    items = ElementList([item
                                if isinstance(item, _base_types.Element)
                                else TextBlock(item)
                                for item in items])
            else:
                items = ElementList([TextBlock(items)])
            __value = items
        elif __name == "min_height":
            min_height = __value
            try:
                min_height = convert_to_pixel_string(min_height)
            except ValueError:
                raise_invalid_pixel_error(__name, min_height)
            __value = min_height
        return super().__setattr__(__name, __value)


class TableRow(Mixin):
    __slots__ = ("type", "cells", "style")
    def __init__(self,
                 cells: ListLike[Any] = DefaultNone,
                 style: Literal["default", "emphasis", "good", "attention", "warning",
                                "accent"] = DefaultNone):
        self.type = "TableRow"
        if cells is DefaultNone:
            cells = ElementList()
        self.cells = cells
        self.style = style
    
    def append(self, __object: _base_types.Element | Any, /):
        if not isinstance(__object, TableCell): __object = TableCell(__object)
        self.cells.append(__object)

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, __k):
        if isinstance(__k, slice):
            return self.__class__(cells=self.cells[__k])
        return self.cells.__getitem__(__k)
    
    def __setitem__(self, __key, __value):
        if not isinstance(__value, TableCell): __value = TableCell(__value)
        self.cells.__setitem__(__key, __value)

    def __contains__(self, value):
        return value in self.cells
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "cells":
            cells = __value
            if not isinstance(cells, ListLike):
                raise TypeError
            cells = ElementList([TableCell(cell) if not isinstance(cell, TableCell) else cell
                           for cell in cells])
            __value = cells
        return super().__setattr__(__name, __value)

    def __len__(self):
        return len(self.cells)

    def __iter__(self):
        return iter(self.cells)
    
    def index(self, value=1, start: int = 0, stop: int = maxsize, /):
        return self.cells.index(value, start, stop)

    def count(self, value, /):
        return self.cells.count(value)
    
    def __reversed__(self):
        attrs = {attr_name: getattr(self, attr_name)
                 if attr_name != "cells"
                 else reversed(getattr(self, attr_name))
                 for attr_name in self.__slots__
                 if hasattr(self, attr_name)}
        return self.__class__(**attrs)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.cells)})"

    def __str__(self):
        return "[" + ", ".join([str(cell) for cell in self.cells]) + "]"


class Table(Mixin):
    __slots__ = ('type', 'columns', 'rows', 'first_row_as_header', 'show_grid_lines', 'grid_style',
                 'horizontal_cell_content_alignment', 'vertical_cell_content_alignment', 'fallback', 'height',
                 'separator', 'spacing', 'id', 'is_visible')
    def __init__(self,
                 rows: ListLike[ListLike] = DefaultNone,
                 first_row_as_header: bool = DefaultNone,
                 columns: ListLike[int] = DefaultNone,
                 show_grid_lines: bool = DefaultNone,
                 grid_style: Literal["default", "emphasis", "good", "attention", "warning",
                                     "accent"] = DefaultNone,
                 horizontal_cell_content_alignment: Literal["left", "center", "right"] = DefaultNone,
                 vertical_cell_content_alignment: Literal["top", "center", "bottom"] = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone):

        self.type = "Table"
        if rows is DefaultNone:
            rows = []
        self.rows: RowList = rows
        self.columns = columns if columns is not None else []
        self.first_row_as_header = first_row_as_header
        self.show_grid_lines = show_grid_lines
        self.grid_style = grid_style
        self.horizontal_cell_content_alignment = horizontal_cell_content_alignment
        self.vertical_cell_content_alignment = vertical_cell_content_alignment
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, __k, /):
        return self.rows.__getitem__(__k)
    
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, TableRow):
            raise TypeError
        return self.rows.__setitem__(__key, __value)
    
    def append(self, row: ListLike):
        if not isinstance(row, ListLike):
            raise TypeError
        if not isinstance(row, TableRow):
            row = TableRow(row)
        self.rows.append(row)
    
    # custom to_dict para lidar com o formato atípico do atributo columns dentro do json
    def to_dict(self):
        dic = super().to_dict()
        if hasattr(self, "columns") and self.columns:
            json_columns = [{"width": width} for width in self.columns]
        elif self.rows:
            json_columns = [{"width": 1} for _ in self.rows[0]]
        else:
            json_columns = dic["columns"]
        dic["columns"] = json_columns
        return dic
    
    def __len__(self):
        return len(self.rows)
    
    def __str__(self):
        rows = [["\n".join([str(item) for item in cell.items]) for cell in row.cells] for row in self.rows]
        if getattr(self, "first_row_as_header", False) and self.rows:
            headers, *rows = rows
            return tabulate(rows, headers, tablefmt='grid')

        return tabulate(rows, tablefmt='grid')

    def __iter__(self):
        return iter(self.rows)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "rows":
            rows = __value
            if not isinstance(rows, ListLike) or \
            not all([isinstance(item, ListLike) for item in rows]):
                raise TypeError("argument 'rows' must be a collection of collections")
            rows = RowList([TableRow(row) for row in rows])
            __value = rows
        return super().__setattr__(__name, __value)


class ActionSet(Mixin):
    __slots__ = ("type", "actions", "fallback", "height", "separator", "spacing", "id", "is_visible")
    def __init__(self,
                 actions: _base_types.Action | ListLike[_base_types.Action] = DefaultNone,
                 fallback: _base_types.Element = DefaultNone,
                 height: Literal["auto", "stretch"] = DefaultNone,
                 separator: bool = DefaultNone,
                 spacing: Literal["default", "none", "small", "medium", "large", "extraLarge",
                                  "padding"] | None = DefaultNone,
                 id: str = DefaultNone,
                 is_visible: bool = DefaultNone
                 ):
        self.type = "ActionSet"
        if actions is DefaultNone:
            actions = ElementList()
        self.actions: ElementList = actions
        self.fallback = fallback
        self.height = height
        self.separator = separator
        self.spacing = spacing
        self.id = id
        self.is_visible = is_visible

    @overload
    def __getitem__(self, __i: int):
        ...

    @overload
    def __getitem__(self, __s: slice):
        ...

    def __getitem__(self, __k):
        r = self.actions.__getitem__(__k)  
        if isinstance(__k, slice):
            r = self.__class__(r)
        return r

    def append(self, action: _base_types.Action) -> None:
        if not isinstance(action, _base_types.Action):
            raise TypeError("Can only append objetcs of type Action")
        self.actions.append(action)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "actions":
            if isinstance(__value, _base_types.Action):
                __value = ElementList([__value])
            elif isinstance(__value, ListLike):
                __value = ElementList(__value)
        return super().__setattr__(__name, __value)
        

_base_types.Container.register(Container)
_base_types.Column.register(Column)
_base_types.ColumnSet.register(ColumnSet)
_base_types.Table.register(Table)
_base_types.TableRow.register(TableRow)
_base_types.TableCell.register(TableCell)
_base_types.ActionSet.register(ActionSet)


from adaptivecard import AdaptiveCard
from adaptivecard.containers import Container, ColumnSet, Column, Table, TableRow
from adaptivecard.card_elements import TextBlock, Image
from adaptivecard.inputs import Text, Time, ChoiceSet, Choice
from adaptivecard.types import Backgroundimage

class Test:
    def test_message(self):
        card = AdaptiveCard()
        container = Container(style='warning', bleed=True, min_height=10)
        columns = [Column([1,1,1]), Column([2,2,2])]
        column_set = ColumnSet((columns), style='warning', min_height=8, height='auto')
        bg_image = Backgroundimage(
            url="https://images.ctfassets.net/sfnkq8lmu5d7/4Ma58uke8SXDQLWYefWiIt/3f1945422ea07ea6520c7aae39180101/2021-11-24_Singleton_Puppy_Syndrome_One_Puppy_Litter.jpg?w=1000&h=750&fl=progressive&q=70&fm=jpg",
            fill_mode="cover",
            horizontal_alignment="center"
        )
        column_set[0].background_image = bg_image
        container.append(column_set)
        card.append(container)
        container_2 = Container()
        table = Table(first_row_as_header=True)
        table.append([1, 2, TextBlock("header", style='heading', weight='bolder')])
        row_2 = TableRow([4,5,6])
        table.append(row_2)
        assert table[1][0][0].text == "4"
        table[1][0] = 100
        assert table[1][0][0].text == "100"
        container_2.append(table)
        card.append(container_2)
        img = Image("https://hips.hearstapps.com/hmg-prod/images/dog-puppy-on-garden-royalty-free-image-1586966191.jpg?crop=1xw:0.74975xh;center,top&resize=1200:*",
                    size='medium', style='person', separator=True, height=70)
        card.append(img)
        text_input = Text(id="text input id")
        time_input = Time(id="time id")
        card.append(text_input)
        card.append(time_input)
        choice_set = ChoiceSet(id="choice set id")
        choice = Choice(title="some title", value="value1")
        choice_set.append(choice)
        card.append(choice_set)
        msg = card.to_message()
        json = msg.to_dict()
        assert isinstance(json, dict)


from setuptools import setup, find_packages

VERSION = '0.3.0'
DESCRIPTION = 'Microsoft Adaptive Cards'
LONG_DESCRIPTION = 'A package that helps you design adaptive cards in an object-oriented manner.'


with open("requirements.txt") as f:
    requires = f.read().split()

setup(
    name="adaptivecard",
    version=VERSION,
    author="cabutchei (Luan Paz)",
    author_email="<luropa_paz@hotmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=requires,
    keywords=['python', 'adaptive', 'cards', 'microsoft'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License", 
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


from typing import Protocol, Sequence, Iterator, Iterable, TypeVar, Any, SupportsIndex, overload, runtime_checkable
from adaptivecard._base_types import Element, Choice, TableRow, Column


_T_co = TypeVar("_T_co", covariant=True)

@runtime_checkable
class SequenceNotStr(Protocol[_T_co]):
    @overload
    def __getitem__(self, index: SupportsIndex, /) -> _T_co:
        ...

    @overload
    def __getitem__(self, index: slice, /) -> Sequence[_T_co]:
        ...

    def __contains__(self, value: object, /) -> bool:
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[_T_co]:
        ...

    def index(self, value: Any, start: int = ..., stop: int = ..., /) -> int:
        ...

    def count(self, value: Any, /) -> int:
        ...

    def __reversed__(self) -> Iterator[_T_co]:
        ...

ListLike = SequenceNotStr

DefaultNone = object()


class ElementList(list):
    def __init__(self, data=None) -> None:
        if data is None: data = []
        super().__init__(data)
    def append(self, __object) -> None:
        if not isinstance(__object, Element):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Element.__name__}")
        return super().append(__object)
    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        if not isinstance(__object, Element):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Element.__name__}")
        return super().insert(__index, __object)
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, Element):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Element.__name__}")
        return super().__setitem__(__key, __value)
    @overload
    def __getitem__(self, __i: int):
        ...
    @overload
    def __getitem__(self, __s: slice):
        ...
    def __getitem__(self, k):
        r = super().__getitem__(k)
        if isinstance(r, list):
            r = self.__class__(r)
        return r


class ColumnList(list):
    def __init__(self, data=None) -> None:
        if data is None: data = []
        super().__init__(data)
    def append(self, __object) -> None:
        if not isinstance(__object, Column):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Column.__name__}")
        return super().append(__object)
    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        if not isinstance(__object, Column):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Column.__name__}")
        return super().insert(__index, __object)
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, Column):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Column.__name__}")
        return super().__setitem__(__key, __value)
    @overload
    def __getitem__(self, __i: int):
        ...
    @overload
    def __getitem__(self, __s: slice):
        ...
    def __getitem__(self, k):
        r = super().__getitem__(k)
        if isinstance(r, list):
            r = self.__class__(r)
        return r


class RowList(list):
    def __init__(self, data=None) -> None:
        if data is None: data = []
        super().__init__(data)
    def append(self, __object) -> None:
        if not isinstance(__object, TableRow):
            raise TypeError(f"{type(self).__name__} only accepts items of type {TableRow.__name__}")
        return super().append(__object)
    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        if not isinstance(__object, TableRow):
            raise TypeError(f"{type(self).__name__} only accepts items of type {TableRow.__name__}")
        return super().insert(__index, __object)
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, TableRow):
            raise TypeError(f"{type(self).__name__} only accepts items of type {TableRow.__name__}")
        return super().__setitem__(__key, __value)
    @overload
    def __getitem__(self, __i: int):
        ...
    @overload
    def __getitem__(self, __s: slice):
        ...
    def __getitem__(self, k):
        r = super().__getitem__(k)
        if isinstance(r, list):
            r = self.__class__(r)
        return r


class ChoiceList(list):
    def __init__(self, data=None) -> None:
        if data is None: data = []
        super().__init__(data)
    def append(self, __object) -> None:
        if not isinstance(__object, Choice):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Choice.__name__}")
        return super().append(__object)
    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        if not isinstance(__object, Choice):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Choice.__name__}")
        return super().insert(__index, __object)
    def __setitem__(self, __key, __value, /):
        if not isinstance(__value, Choice):
            raise TypeError(f"{type(self).__name__} only accepts items of type {Choice.__name__}")
        return super().__setitem__(__key, __value)
    @overload
    def __getitem__(self, __i: int):
        ...
    @overload
    def __getitem__(self, __s: slice):
        ...
    def __getitem__(self, k):
        r = super().__getitem__(k)
        if isinstance(r, list):
            r = self.__class__(r)
        return r


from typing import Literal
from adaptivecard._mixin import Mixin
import adaptivecard._base_types as _base_types
from adaptivecard._typing import ListLike, DefaultNone


class ShowCard(Mixin):
    __slots__ = ('type', 'title', 'icon_url', 'id', 'style', 'fallback', 'tooltip', 'is_enabled',
                 'mode', 'card')
    def __init__(self,
                 card: _base_types.AdaptiveCard,
                 title: str = DefaultNone,
                 icon_url: str = DefaultNone,
                 id: str = DefaultNone,
                 style: Literal["default", "positive", "destructive"] = DefaultNone,
                 fallback: _base_types.Action = DefaultNone,
                 tooltip: str = DefaultNone,
                 is_enabled: bool = DefaultNone,
                 mode: Literal["primary", "secondary"] = DefaultNone):
        
        self.type = "Action.ShowCard"
        self.title = title
        self.icon_url = icon_url
        self.id = id
        self.style = style
        self.fallback = fallback
        self.tooltip = tooltip
        self.is_enabled = is_enabled
        self.mode = mode
        self.card = card


class OpenUrl(Mixin):
    __slots__ = ("url", "title", "id", "style", "fallback", "tooltip", "is_enabled", "mode")
    def __init__(self,
                 url: str,
                 title: str = DefaultNone,
                 id: str = DefaultNone,
                 style: Literal["default", "positive", "destructive"] = DefaultNone,
                 fallback: _base_types.Action  = DefaultNone,
                 tooltip: str = DefaultNone,
                 is_enabled: bool = DefaultNone,
                 mode: Literal["primary", "secondary"] = DefaultNone):

        self.type = "Action.OpenUrl"
        self.title = title
        self.url = url
        self.id = id
        self.style = style
        self.fallback = fallback
        self.tooltip = tooltip
        self.is_enabled = is_enabled
        self.mode = mode


class Submit(Mixin):
    __slots__ = ("type", "data", "associated_inputs", "title", "icon_url", "id", "style", "fallback",
                 "tooltip", "is_enabled", "mode")
    def __init__(self,
                 data: dict,
                 associated_inputs: Literal["auto", "none"] | None,
                 title: str = DefaultNone,
                 icon_url: str = DefaultNone,
                 id: str = DefaultNone,
                 style: Literal["default", "positive", "destructive"] = DefaultNone,
                 fallback: _base_types.Action = DefaultNone,
                 tooltip: str = DefaultNone,
                 is_enabled: bool = DefaultNone,
                 mode: Literal["primary", "secondary"] = DefaultNone):

        self.type = "Action.ToggleVisibility"
        self.data = data
        self.associated_inputs = associated_inputs
        self.title = title
        self.icon_url = icon_url
        self.id = id
        self.style = style
        self.fallback = fallback
        self.tooltip = tooltip
        self.is_enabled = is_enabled
        self.mode = mode


class TargetElement(Mixin):
    __slots__ = ("element_id", "is_visible")
    def __init__(self,
                 element_id: str,
                 is_visible: bool):
        self.element_id = element_id
        self.is_visible = is_visible


class ToggleVisibilty(Mixin):
    __slots__ = ("type", "data", "target_elements", "icon_url", "title", "id", "style", "fallback",
                 "tooltip", "is_enabled", "mode")
    def __init__(self,
                 data: str | dict,
                 target_elements: ListLike[TargetElement],
                 icon_url: str = DefaultNone,
                 title: str = DefaultNone,
                 id: str = DefaultNone,
                 style: Literal["default", "positive", "destructive"] = DefaultNone,
                 fallback: _base_types.Action = DefaultNone,
                 tooltip: str = DefaultNone,
                 is_enabled: bool = DefaultNone,
                 mode: Literal["primary", "secondary"] = DefaultNone):

        self.type = "Input.ToggleVisibility"
        self.data = data
        self.target_elements = target_elements
        self.icon_url = icon_url
        self.title = title
        self.id = id
        self.style = style
        self.fallback = fallback
        self.tooltip = tooltip
        self.is_enabled = is_enabled
        self.mode = mode

    
class Execute(Mixin):
    __slots__ = ()
    def __init__(self,
                 verb: str = DefaultNone,
                 data: dict = DefaultNone,
                 associated_inputs: Literal["auto", "none"] | None = DefaultNone,
                 title: str = DefaultNone,
                 icon_url: str = DefaultNone,
                 id: str = DefaultNone,
                 style: Literal["default", "positive", "destructive"] = DefaultNone,
                 fallback: _base_types.Action = DefaultNone,
                 tooltip: str = DefaultNone,
                 is_enabled: bool = DefaultNone,
                 mode: Literal["primary", "secondary"] = DefaultNone):
        
        self.type = "Input.Execute"
        self.verb = verb
        self.data = data
        self.associated_inputs = associated_inputs
        self.title = title
        self.icon_url = icon_url
        self.id = id
        self.style = style
        self.fallback = fallback
        self.tooltip = tooltip
        self.is_enabled = is_enabled
        self.mode = mode


_base_types.ShowCard.register(ShowCard)
_base_types.OpenUrl.register(OpenUrl)
_base_types.Submit.register(Submit)
_base_types.ToggleVisibility.register(ToggleVisibilty)
_base_types.Execute.register(Execute)